# pylint: disable=c0111

import asyncio

from unittest import TestCase

from asynciojobs import Scheduler, Job, Sequence

from .util import produce_svg


VERBOSE = False

def verbose(*args, **kwds):
    if not VERBOSE:
        return
    print(*args, **kwds)


async def aprint(*messages):
    print(messages)


class CounterScheduler(Scheduler):
    def __init__(self, *a, **k):
        super().__init__(*a, verbose=VERBOSE, **k)
        self.counter = 0


class CounterJob(Job):
    def __init__(self, scheduler, delay, shutdown_delay, *args, **kwds):
        """
        scheduler is **NOT** the scheduler where the job belongs directly
          but instead the global toplevel scheduler. This way all the jobs
          can make side effect on the counter attached there.
        delay is, in tens of seconds, the time that co_shutdown will take

        example: delay=15 -> shutdown takes 1.5 s
        """
        self.scheduler = scheduler
        self.delay = delay
        self.shutdown_delay = shutdown_delay
        super().__init__(*args, **kwds)

    async def co_run(self):
        verbose(f">>> CounterJob.run {self.label}")
        await asyncio.sleep(self.delay)
        self.scheduler.counter += 1
        verbose(f"<<< CounterJob.run {self.label}")

    async def co_shutdown(self):
        # delay in tens of seconds
        verbose(f">>> CounterJob.shutdown {self.label}")
        if not self.is_done():
            verbose(f"ignoring job {self.label} - is_done={self.is_done()}")
            return
        await asyncio.sleep(self.shutdown_delay)
        self.scheduler.counter -= 1
        verbose(f"<<< CounterJob.shutdown {self.label}")


class Tests(TestCase):

    # simple / sequential scheduler

    def _test_simple(self, sched_timeout):

        cardinal = 4

        sched = CounterScheduler(critical=False, timeout=sched_timeout)
        sched.add(Sequence(
            *[CounterJob(sched, delay=(i+1)/10, shutdown_delay=0,
                         corun=aprint(i),
                         label=f"simple {i}",
                         )
              for i in range(cardinal)]))

        self.assertEqual(sched.counter, 0)
        sched.run()
# shutdown is now implicit
#        self.assertEqual(sched.counter, cardinal)
#        self.assertTrue(sched.shutdown())
        self.assertEqual(sched.counter, 0)
        sched.close_all_jobs()

    def test_simple(self):
        self._test_simple(sched_timeout=None)

    def test_simple_timeout(self):
        self._test_simple(sched_timeout=0.5)


    # mostly the same with a nested scheduler

    def _test_nested(self, sched_timeout):

        cardinal = 3

        # same to the square
        top = CounterScheduler(critical=False, timeout=sched_timeout,
                               label="OUT")
        subs = []
        for i in range(cardinal):
            sub = Scheduler(label=f"SUB {i}")
            subs.append(sub)
            sub.add(Sequence(
                *[CounterJob(top, delay=(i+j)/10, shutdown_delay=0,
                             corun=aprint('ok'),
                             label=f"nested {i} x {j}")
                  for j in range(cardinal)]))
        top.add(Sequence(*subs))

        self.assertEqual(top.counter, 0)
        top.run()
# shutdown is now implicit
#        self.assertEqual(top.counter, cardinal*cardinal)
#        self.assertTrue(top.shutdown())
        self.assertEqual(top.counter, 0)

        top.close_all_jobs()

    def test_nested(self):
        self._test_nested(sched_timeout=None)

    def test_nested_timeout(self):
        self._test_nested(sched_timeout=0.5)


    # with a shutdown_timeout
    # much like the simple version but
    #
    # we create 2 pipes of <cardinal> jobs that all take 0.1s
    #
    # one of the pipes has all its jobs' shutdown() take 0.1s
    # the other has all its jobs' shutdown()        take 0.5s

    def _test_shutdown_timeout(self, sched_timeout, shutdown_timeout):

        cardinal = 4
        run_delay = 0.1
        delays = (0.1, 0.5)

        sched = CounterScheduler(critical=False,
                                 timeout=sched_timeout,
                                 shutdown_timeout=shutdown_timeout)
        for shutdown_delay in delays:
            sched.add(Sequence(
                *[CounterJob(sched, delay=run_delay,
                             shutdown_delay=shutdown_delay,
                             corun=aprint(i),
                             label=f"sched-timeout {i} x {shutdown_delay}",
                             )
                  for i in range(cardinal)]))

        self.assertEqual(sched.counter, 0)
        sched.run()

        ## compute expected remaining counter
        expected = 0
        if not sched_timeout:
            actual = cardinal
        elif sched_timeout == 0.25:
            # shutdown works only on jobs that are done,
            # so at this point we're in the middle of the third row of jobs
            # and only 2 jobs per pipe have completed
            actual = 2
        else:
            print("INTERNAL ERROR")
        ## with no execution timeout
        # all jobs have run, result will depend on
        # how shutdown_timeout compares with short and long
        short, long = delays
        if shutdown_timeout < short:
            expected = 2 * actual
        elif short <= shutdown_timeout < long:
            expected = actual
        else:
            expected = 0
        produce_svg(sched, "debug")

        self.assertEqual(sched.counter, expected)

        sched.close_all_jobs()

    def test_shsd_none_short(self):
        return self._test_shutdown_timeout(None, 0.05)
    def test_shsd_none_medium(self):
        return self._test_shutdown_timeout(None, 0.2)
    def test_shsd_none_long(self):
        return self._test_shutdown_timeout(None, 0.7)

    def test_shsd_aborted_short(self):
        return self._test_shutdown_timeout(0.25, 0.05)
    def test_shsd_aborted_medium(self):
        return self._test_shutdown_timeout(0.25, 0.2)
    def test_shsd_aborted_long(self):
        return self._test_shutdown_timeout(0.25, 0.7)



"""
The ``Scheduler`` class makes it easier to nest scheduler objects.
"""

# pylint: disable=w0212

from asynciojobs import PureScheduler
from asynciojobs import AbstractJob


class Scheduler(PureScheduler, AbstractJob):
    """
    The ``Scheduler`` class is a mixin of the two
    :class:`~asynciojobs.purescheduler.PureScheduler` and
    :class:`~asynciojobs.job.AbstractJob` classes.

    As such it can be used to create nested schedulers,
    since it is a scheduler that can contain jobs,
    and at the same time it is a job, and so it can be included in
    a scheduler.

    Parameters:
      jobs_or_sequences: passed to
        :class:`~asynciojobs.purescheduler.PureScheduler`,
        allows to add these jobs inside of the newly-created scheduler;
      jobs_window: passed to
        :class:`~asynciojobs.purescheduler.PureScheduler`;
      timeout: passed to
        :class:`~asynciojobs.purescheduler.PureScheduler`;
      shutdown_timeout: passed to
        :class:`~asynciojobs.purescheduler.PureScheduler`;
      watch (Watch): passed to
        :class:`~asynciojobs.purescheduler.PureScheduler`;
      verbose (bool): passed to
        :class:`~asynciojobs.purescheduler.PureScheduler`;
      kwds: all other named arguments are sent
        to the :class:`~asynciojobs.job.AbstractJob` constructor.


    Example:

      Here's how to create a very simple scheduler with
      an embedded sub-scheduler; the whole result is equivalent to a simple
      4-steps sequence::

        main = Scheduler(
           Sequence(
             Job(aprint("begin", duration=0.25)),
             Scheduler(
               Sequence(
                 Job(aprint("middle-begin", duration = 0.25)),
                 Job(aprint("middle-end", duration = 0.25)),
               )
             ),
             Job(aprint("end", duration=0.25)),
           )
        main.run()

    Notes:

      There can be several good reasons for using nested schedulers:

      * the scope of a ``window`` object applies to a scheduler, so a nested
        scheduler is a means to apply windoing on a specific set of jobs;
      * likewise the ``timeout`` attribute only applies to the run for the
        whole scheduler;
      * you can use ``forever`` jobs that will be terminated earlier than
        the end of the global scheduler;
      * strictly speaking, the outermost instance in this example could be an
        instance of ``PureScheduler``, but in practice it is simpler to always
        create instances of ``Scheduler``.

      Using an intermediate-level scheduler can in some case help alleviate or
      solve such issues.

    """
    def __init__(self, *jobs_or_sequences,
                 jobs_window=None, timeout=None,
                 shutdown_timeout=1,
                 watch=None, verbose=False,
                 **kwds):

        PureScheduler.__init__(self, *jobs_or_sequences,
                               jobs_window=jobs_window, timeout=timeout,
                               shutdown_timeout=shutdown_timeout,
                               watch=watch, verbose=verbose)
        AbstractJob.__init__(self, **kwds)

    async def co_run(self):
        """
        Supersedes the :meth:`~asynciojobs.puresheduler.PureScheduler.co_run`
        method in order to account for **critical** schedulers.

        `Scheduler` being a subclass of `AbstractJob`, we need to account
        for the possibility that a scheduler is defined as ``critical``.

        If the inherited ``co_run()`` method fails because
        of an exception of a timeout, a critical Scheduler will trigger an
        exception, instead of returning ``False``:

        * if orchestration failed because an internal job has raised an
          exception, raise that exception;
        * if it failed because of a timeout, raise ``TimeoutError``

        Returns:
          bool: ``True`` if everything went well;
            ``False`` for non-critical schedulers that go south.

        Raises:
          TimeoutError: for critical schedulers that do not complete in time,
          Exception: for a critical scheduler that has a critical job that
            triggers an exception, in which case it bubbles up.
        """
        # run as a pure scheduler, will always return True or False
        pure = await PureScheduler.co_run(self)
        # fine
        if pure is True:
            return pure
        # non-critical : we're done
        if not self.critical:
            return pure
        # a timeout
        if self.failed_time_out():
            raise TimeoutError("critical scheduler took too long")
        # a critical job has exploded
        if self.failed_critical():
            # need to find at least one critical job
            # that has raised an exception
            for job in self.jobs:
                if not job.critical:
                    continue
                exc = job.raised_exception()
                if exc:
                    raise exc
        # we should not reach this point
        raise ValueError("Internal error in Scheduler.co_run()")

    def _set_sched_id(self, start, id_format):
        """
        Works as a complicit to PureScheduler._set_sched_ids.
        It sets local the ``_sched_id`` attribute and returns the index
        for the next job.
        """
        # first set index on the current (kind of fake) node
        i = AbstractJob._set_sched_id(self,
                                      start, id_format)
        # go on with the jobs in sub scheduler
        return PureScheduler._set_sched_ids(self,
                                            i, id_format)

    def _job_count(self):
        """
        Complicit to PureScheduler._total_length()

        Returns:
          int: 1 + number of nodes included/nested
        """
        return 1 + self._total_length()

    def _list(self, details, depth, recursive, silence_done_jobs):
        """
        Complicit to PureScheduler.list()
        """
        indent = ('>'*depth + ' ') if depth else ''
        print(f"{self.repr_id()} {self.repr_short()} {indent}{self.repr_main()} "
              f"{self.repr_result()} {self.repr_requires()} -> {self.repr_entries()}")
        if recursive:
            for job in self.topological_order():
                job._list(details, depth+1, recursive, silence_done_jobs)
            print(self.repr_id(),
                  # this should be 7-spaces like repr_short()
                  '--end--',
                  '<'*(depth+1),
                  self.repr_main(),
                  self.repr_exits())

    def _list_safe(self, recursive, silence_done_jobs):
        """
        Complicit to PureScheduler.list_safe()
        """
        print(f"{self.repr_short()} {self.repr_id()} {self.repr_main()} {self.repr_requires()}")
        if recursive:
            for job in self.jobs:
                job._list_safe(recursive, silence_done_jobs=silence_done_jobs)
            print('--end--', self.repr_id())

    def _iterate_jobs(self, scan_schedulers):
        if scan_schedulers:
            yield self
        for job in self.jobs:
            yield from job._iterate_jobs(
                scan_schedulers=scan_schedulers)

    def dot_cluster_name(self):
        """
        assigns a name to the subgraph that will represent
        a nested scheduler; dot format imposes this name to start
        with ``cluster_``
        """
        return f"cluster_{self._sched_id}"

    def check_cycles(self):
        """
        Supersedes
        :meth:`~asynciojobs.puresheduler.PureScheduler.check_cycles`
        to account for nested schedulers.

        Returns:
          bool: True if this scheduler, and all its nested schedulers
            at any depth, has no cycle and can be safely scheduled.
        """
        try:
            for job in self.topological_order():
                if isinstance(job, Scheduler) and not job.check_cycles():
                    return False
            return True
        except Exception as exc:                        # pylint: disable=W0703
            if self.verbose:
                print("check_cycles failed", exc)
            return False



"""
Schedulers and jobs requirements are essentially sets of jobs,
and from a semantic point of view, order does not matter.

However for debugging/cosmetic reasons,
keeping track of creation order can be convenient.

So using ``OrderedSet`` looks like a good idea;
but it turns out that on some distros like fedora,
installing ``OrderedSet`` can be a pain, as it involves recompiling C code,
which in turn pulls in a great deal of dependencies.

For this reason, we use ``OrderedSet`` only if available,
and resort to regular sets otherwise.
On macos or ubuntu, fortunately, this can be simply achieved with::

    pip3 install orderedset

or alternatively with::

    pip3 install asynciojobs[ordered]
"""

# pylint: disable=e0401, w0611

try:
    from orderedset import OrderedSet as BestSet
except ImportError:
    BestSet = set



# pylint: disable=c0111, c0103, r0201

import unittest

from asynciojobs import PureScheduler, Scheduler, Job, Sequence, Watch

from .util import co_print_sleep


class Tests(unittest.TestCase):

    def test_flat_cycles(self):

        watch = Watch()

        def job_in_s(i, s):
            return Job(co_print_sleep(watch, .2, f"job {i}"),
                       label=f"job{i}",
                       scheduler=s)

        def check_class(klass):
            sched = klass()
            j1, j2, j3 = [job_in_s(i, sched) for i in range(1, 4)]
            j3.requires(j2)
            j2.requires(j1)

            # no cycle yet
            self.assertTrue(sched.check_cycles())

            # create cycle
            j1.requires(j3)
            self.assertFalse(sched.check_cycles())
            sched.close_idle_jobs()

        check_class(PureScheduler)
        check_class(Scheduler)

    def test_nested_cycles(self):

        watch = Watch()

        def job(i):
            return Job(co_print_sleep(watch, .2, f"job {i}"),
                       label=f"job{i}")
        js1, js2, js3 = [job(i) for i in range(11, 14)]
        s2 = Scheduler(Sequence(js1, js2, js3))

        j1, j3 = job(1), job(3)
        s1 = Scheduler(Sequence(j1, s2, j3))
        self.assertTrue(s1.check_cycles())

        # create cycle in subgraph
        js1.requires(js3)
        self.assertFalse(s1.check_cycles())

        # restore in OK state
        js1.requires(js3, remove=True)
        self.assertTrue(s1.check_cycles())

        # add cycle in toplevel
        j1.requires(j3)
        self.assertFalse(s1.check_cycles())

        # restore in OK state
        j1.requires(j3, remove=True)
        self.assertTrue(s1.check_cycles())

        # add one level down
        s3 = Scheduler()
        jss1, jss2, jss3 = [job(i) for i in range(111, 114)]
        Sequence(jss1, jss2, jss3, scheduler=s3)

        # surgery in s2; no cycles
        s2.remove(js2)
        s2.sanitize()
        s2.add(s3)
        s3.requires(js1)
        js3.requires(s3)
        self.assertTrue(s1.check_cycles())

        # add cycle in s3
        js1.requires(js3)
        self.assertFalse(s1.check_cycles())

        for s in (s1, s2, s3):
            s.close_idle_jobs()
        # this one has been removed manually
        js2.close()



# pylint: disable=c0111, c0103

import unittest

from asynciojobs import PureScheduler, Job, Sequence, Watch

from .util import co_print_sleep, produce_svg


class Tests(unittest.TestCase):

    def test_graph1(self):

        watch = Watch()

        s = PureScheduler()
        s.add(Sequence(
            Job(co_print_sleep(watch, .25, 'begin')),
            Job(co_print_sleep(watch, 1., 'middle'), label='middle'),
            Job(co_print_sleep(watch, .25, 'end')),
        ))
        print("test_graph1 NO DETAILS")
        s.list()
        print("test_graph1 WITH DETAILS")
        s.list(details=True)
        print("GRAPH")
        self.assertEqual(len(s), 3)
        s.run()
        self.assertAlmostEqual(watch.seconds(), 1.5, delta=0.05)
        produce_svg(s, "graphic-graph1")

    def test_graph2(self):

        class TextJob(Job):
            def __init__(self, text, *args, **kwds):
                self.text = text
                super().__init__(*args, **kwds)

            def text_label(self):
                return "[[TextJob {}]]".format(self.text[::-1])

            def details(self):
                return f"TextJob details say\n" \
                    f"that initial text was {self.text}"

        class GraphJob(Job):
            def __init__(self, graph, *args, **kwds):
                self.graph = graph
                super().__init__(*args, **kwds)

            def graph_label(self):
                return "[[GraphJob\n{}]]".format(self.graph[::-1])

            def details(self):
                return f"GraphJob details\nare even more verbose and say\n" \
                       f"that initial graph message\nwas {self.graph}"

        watch = Watch()
        sched = PureScheduler()
        sched.add(Sequence(
            TextJob('textjob-with',
                    co_print_sleep(watch, 0.1, 'textjob, no label ')),
            TextJob('textjob-without',
                    co_print_sleep(watch, 0.1, 'textjob, with label '),
                    label='TextLabel'),
            GraphJob('graphjob-with',
                     co_print_sleep(watch, 0.1, 'graphjob, no label ')),
            GraphJob('graphjob-without',
                     co_print_sleep(watch, 0.1, 'graphjob, with label '),
                     label='GraphLabel'),
        ))
        print("test_graph2 NO DETAILS")
        sched.list()
        print("test_graph2 WITH DETAILS")
        sched.list(details=True)
        print("GRAPH")
        self.assertEqual(len(sched), 4)
        produce_svg(sched, "graphic-graph2")

        sched.close_idle_jobs()



# pylint: disable=c0111, c0103, r0201

from re import M
import unittest

from asynciojobs import *
from .util import *

def job_in_s_w(sched, w, i):
    return Job(
        co_print_sleep(w, .2, f"job {i}"),
        label=f"job{i}",
        scheduler=sched)

# cannot safely use sched.close_done_jobs() nor close_all_jobs()
# because in these tests we keep on creating fake jobs
# that no longer belong in the scheduler after various
# bypass manipulations

class Tests(unittest.TestCase):

    def test_bypass1(self):

        w = Watch()
        sched = Scheduler()

        jobs = j1, j2, j3, j4, j5 = [job_in_s_w(sched, w, i) for i in range(1, 6)]
        j5.requires(
            j4.requires(
                j3.requires(
                    j2.requires(j1))))
        self.assertEqual(sched.successors_downstream(j1), {j2, j3, j4, j5})
        self.assertEqual(sched.predecessors_upstream(j5), {j1, j2, j3, j4})
        produce_svg(sched, "graphic-bypass1-step0")


        sched.bypass_and_remove(j3)
        self.assertTrue(sched.sanitize())
        self.assertFalse(j3 in sched.jobs)
        self.assertTrue(j2 in j4.required)
        produce_svg(sched, "graphic-bypass1-step1")

        for j in jobs:
            j.close()

    def test_bypass2(self):

        w = Watch()
        sched = Scheduler()

        jobs = j1, j2, j3, j4, j5 = [job_in_s_w(sched, w, i) for i in range(1, 6)]
        j5.requires(
            j4.requires(
                j3.requires(j1),
                j2.requires(j1)))
        self.assertEqual(sched.successors_downstream(j1), {j2, j3, j4, j5})
        self.assertEqual(sched.successors_downstream(j2), {j4, j5})
        self.assertEqual(sched.successors_downstream(j3), {j4, j5})
        self.assertEqual(sched.predecessors_upstream(j4), {j1, j2, j3})
        produce_svg(sched, "graphic-bypass2-step0")

        sched.bypass_and_remove(j4)
        self.assertTrue(sched.sanitize())
        self.assertFalse(j4 in sched.jobs)
        self.assertTrue(j2 in j5.required)
        self.assertTrue(j3 in j5.required)
        self.assertEqual(sched.successors_downstream(j1), {j2, j3, j5})
        self.assertEqual(sched.predecessors_upstream(j5), {j1, j2, j3})
        produce_svg(sched, "graphic-bypass2-step1")

        for j in jobs:
            j.close()

    def diamonds(self, n):
        r"""
        n should be 2 or 3

        with n=2 this returns a watch, a scheduler with 7 jobs
        j1 - j2 - j4 - j5 - j7
           \ j3 /    \ j6 /
        with n=3 it returns a watch, a scheduler with 10 jobs
        j1 - j2 - j4 - j5 - j7 - j8 - j10
           \ j3 /    \ j6 /    \ j9 /
        """

        w = Watch()
        sched = Scheduler()

        j1, j2, j3, j4, j5, j6, j7 = [job_in_s_w(sched, w, i) for i in range(1, 8)]
        j7.requires(
            j6.requires(j4),
            j5.requires(j4),
        )
        j4.requires(
            j3.requires(j1),
            j2.requires(j1),
        )
        if n == 2:
            return w, sched, j1, j2, j3, j4, j5, j6, j7
        j8, j9, j10 = [job_in_s_w(sched, w, i) for i in range(8, 11)]
        j10.requires(
            j8.requires(j7),
            j9.requires(j7),
        )
        if n == 3:
            return w, sched, j1, j2, j3, j4, j5, j6, j7, j8, j9, j10


    def test_bypass3(self):

        _, _, *jobs = w, sched, j1, j2, j3, j4, j5, j6, j7 = self.diamonds(2)

        self.assertEqual(sched.successors_downstream(j1), {j2, j3, j4, j5, j6, j7})
        self.assertEqual(sched.successors_downstream(j4), {j5, j6, j7})
        self.assertEqual(sched.successors_downstream(j7), set())
        self.assertEqual(sched.predecessors_upstream(j1), set())
        self.assertEqual(sched.predecessors_upstream(j4), {j1, j2, j3})
        self.assertEqual(sched.predecessors_upstream(j7), {j1, j2, j3, j4, j5, j6})
        produce_svg(sched, "graphic-bypass3-step0")

        sched.bypass_and_remove(j4)
        self.assertTrue(sched.sanitize())
        self.assertFalse(j4 in sched.jobs)
        self.assertTrue(j2 in j5.required)
        self.assertTrue(j3 in j5.required)
        self.assertTrue(j2 in j6.required)
        self.assertTrue(j3 in j6.required)
        self.assertEqual(sched.successors_downstream(j1), {j2, j3, j5, j6, j7})
        self.assertEqual(sched.successors_downstream(j3), {j5, j6, j7})
        self.assertEqual(sched.successors_downstream(j6), {j7})
        self.assertEqual(sched.successors_downstream(j7), set())
        self.assertEqual(sched.predecessors_upstream(j1), set())
        self.assertEqual(sched.predecessors_upstream(j3), {j1})
        self.assertEqual(sched.predecessors_upstream(j6), {j1, j2, j3})
        self.assertEqual(sched.predecessors_upstream(j7), {j1, j2, j3, j5, j6})
        produce_svg(sched, "graphic-bypass3-step1")

        sched.bypass_and_remove(j1)
        self.assertTrue(sched.sanitize())
        self.assertTrue(len(j2.required) == 0)
        self.assertTrue(len(j3.required) == 0)
        self.assertEqual(sched.successors_downstream(j3), {j5, j6, j7})
        self.assertEqual(sched.predecessors_upstream(j7), {j2, j3, j5, j6})
        produce_svg(sched, "graphic-bypass3-step2")

        sched.bypass_and_remove(j7)
        self.assertTrue(sched.sanitize())
        self.assertFalse(set(sched.successors(j5)))
        self.assertFalse(set(sched.successors(j6)))
        produce_svg(sched, "graphic-bypass3-step3")

        for j in jobs:
            j.close()


    def test_bypass_seq(self):

        w = Watch()
        sched = Scheduler()

        jobs = j1, j2, j3, j4, j5 = [job_in_s_w(sched, w, i) for i in range(1, 6)]
        Sequence(*jobs, scheduler=sched)

        produce_svg(sched, "graphic-bypass-seq-step0")

        sched.bypass_and_remove(j3)
        self.assertTrue(sched.sanitize())
        self.assertFalse(j3 in sched.jobs)
        self.assertTrue(j2 in j4.required)
        self.assertEqual(sched.successors_downstream(j1), {j2, j4, j5})
        self.assertEqual(sched.successors_downstream(j2), {j4, j5})
        self.assertEqual(sched.successors_downstream(j4), {j5})
        self.assertEqual(sched.successors_downstream(j5), set())
        self.assertEqual(sched.predecessors_upstream(j1), set())
        self.assertEqual(sched.predecessors_upstream(j2), {j1})
        self.assertEqual(sched.predecessors_upstream(j4), {j1, j2})
        self.assertEqual(sched.predecessors_upstream(j5), {j1, j2, j4})
        produce_svg(sched, "graphic-bypass-seq-step1")

        for j in jobs:
            j.close()


    def test_bypass_keep1(self):

        _, _, *jobs = w, sched, j1, j2, j3, j4, j5, j6, j7 = self.diamonds(2)
        produce_svg(sched, "graphic-bypass-keep1-step0")
        sched.keep_only_between(starts=[j2, j3], ends=[j5, j6])
        self.assertEqual(len(sched), 5)
        self.assertEqual(sched.jobs, {j2, j3, j4, j5, j6})
        produce_svg(sched, "graphic-bypass-keep1-step1")

        for j in jobs:
            j.close()

    def test_bypass_keep2(self):

        _, _, *jobs = w, sched, j1, j2, j3, j4, j5, j6, j7 = self.diamonds(2)
        produce_svg(sched, "graphic-bypass-keep2-step0")
        sched.keep_only_between(starts=[j4])
        self.assertEqual(len(sched), 4)
        self.assertEqual(sched.jobs, {j4, j5, j6, j7})
        produce_svg(sched, "graphic-bypass-keep2-step1")

        for j in jobs:
            j.close()

    def test_bypass_keep3(self):

        _, _, *jobs = w, sched, j1, j2, j3, j4, j5, j6, j7 = self.diamonds(2)
        produce_svg(sched, "graphic-bypass-keep3-step0")
        sched.keep_only_between(starts=[j4], keep_starts=False)
        self.assertEqual(sched.jobs, {j5, j6, j7})
        produce_svg(sched, "graphic-bypass-keep3-step1")

        for j in jobs:
            j.close()

    def test_bypass_keep4(self):
        _, _, *jobs = w, sched, j1, j2, j3, j4, j5, j6, j7, j8, j9, j10 = self.diamonds(3)
        j_odd = job_in_s_w(sched, w, 11)
        j7.requires(j_odd.requires(j2))
        produce_svg(sched, "graphic-bypass-keep4-step0")

        sched.keep_only_between(starts=[j2, j3], ends=[j5, j6])
        self.assertEqual(len(sched), 5)
        self.assertEqual(sched.jobs, {j2, j3, j4, j5, j6})
        produce_svg(sched, "graphic-bypass-keep4-step1")

        for j in jobs:
            j.close()

    def test_bypass_keep5(self):
        _, _, *jobs = w, sched, j1, j2, j3, j4, j5, j6, j7, j8, j9, j10 = self.diamonds(3)
        j_odd = job_in_s_w(sched, w, 11)
        j7.requires(j_odd.requires(j2))
        produce_svg(sched, "graphic-bypass-keep5-step0")

        sched.keep_only_between(starts=[j2, j3], ends=[j8, j9], keep_ends=False)
        self.assertEqual(len(sched), 7)
        self.assertEqual(sched.jobs, {j2, j3, j4, j5, j6, j7, j_odd})
        produce_svg(sched, "graphic-bypass-keep5-step1")

        for j in jobs:
            j.close()

    def test_bypass_keep10(self):
        _, _, *jobs = w, sched, j1, j2, j3, j4, j5, j6, j7, j8, j9, j10 = self.diamonds(3)
        sched.keep_only([j1, j2, j3, j4])
        self.assertEqual(len(sched), 4)
        self.assertEqual(len(set(sched.successors(j1))), 2)
        self.assertEqual(len(set(sched.predecessors(j1))), 0)
        self.assertEqual(len(set(sched.successors(j2))), 1)
        self.assertEqual(len(set(sched.predecessors(j2))), 1)
        self.assertEqual(len(set(sched.successors(j4))), 0)
        self.assertEqual(len(set(sched.predecessors(j4))), 2)

        for j in jobs:
            j.close()

    def test_bypass_keep11(self):
        _, _, *jobs = w, sched, j1, j2, j3, j4, j5, j6, j7, j8, j9, j10 = self.diamonds(3)
        sched.keep_only([j1, j4, j7])
        self.assertEqual(len(sched), 3)
        self.assertEqual(len(set(sched.successors(j1))), 0)
        self.assertEqual(len(set(sched.predecessors(j1))), 0)
        self.assertEqual(len(set(sched.successors(j4))), 0)
        self.assertEqual(len(set(sched.predecessors(j4))), 0)
        self.assertEqual(len(set(sched.successors(j7))), 0)
        self.assertEqual(len(set(sched.predecessors(j7))), 0)

        for j in jobs:
            j.close()



"""
the asynciojobs package

https://github.com/parmentelat/asynciojobs
https://asynciojobs.readthedocs.io/

"""

from .version import __version__

from .purescheduler import PureScheduler
from .job import AbstractJob, Job
from .scheduler import Scheduler
from .sequence import Sequence
from .watch import Watch
from .printjob import PrintJob



"""
DotStyle is a helper class that takes care of details
when converting data to the DOT format.

This is not meant to be exported outside of this package.
"""

# pylint: disable=c0111


class DotStyle(dict):

    """
    DotStyle is a collection of formatting helpers for creating a DOT output
    """

    def __repr__(self):
        """
        repr() for a DotStyle instance is designed to be shown in the
        [] section for e.g. a node.
        """

        return ",".join(
            f"{key}={DotStyle.value(value_s)}"
            for key, value_s in self.items())

# we found a trick that removed the need for this
#    def cluster_repr(self):
#        """
#        same, but for being mentioned in a subgraph section, like e.g:
#
#            style = "foo,bar";
#            label = "some label";
#        """
#        return "".join(
#            "{}={};\n".format(key, DotStyle.value(value_s))
#            for key, value_s in self.items())

    @staticmethod
    def protect(data):
        string = str(data)
        # escape any double quote
        result = string.replace('"', r'\"')
        # and put double quotes around all this
        return f'"{result}"'

    # output a list (typically for keys like 'shape'
    # or atomic (typically for 'label') value
    @staticmethod
    def value(value_s):
        if isinstance(value_s, list):
            return DotStyle.protect(format(",".join(v for v in value_s)))
        return DotStyle.protect(value_s)



# pylint: disable=c0111

import asyncio

from pathlib import Path

from asynciojobs import Watch, Job, Scheduler


async def co_print_sleep(watch, duration, message):
    # if start_watch was not called, show time with milliseconds
    if watch is None:
        Watch.print_wall_clock()
    else:
        watch.print_elapsed()
    print(message)
    await asyncio.sleep(duration)
    return f"duration={duration}"


# def produce_png(scheduler, name):
#     produce_graphic(scheduler, name, format="png")

def produce_svg(scheduler, name):
    produce_graphic(scheduler, name, format="svg")

def produce_graphic(scheduler, name, format):

    tests_dir = Path('tests')
    if tests_dir.exists():
        actual_name = tests_dir / name
    else:
        actual_name = Path(name)

    scheduler.export_as_graphic(actual_name, format)
    print(f"graphic files produced in {actual_name}.{{dot,{format}}}")


def diamond_from_jobs(watch, j1, j2, j3, j4, scheduler_class=Scheduler):
    diamond = scheduler_class(watch=watch)
    diamond.add(j1)
    diamond.add(j2)
    diamond.add(j3)
    diamond.add(j4)
    j2.requires(j1)
    j3.requires(j1)
    j4.requires(j2)
    j4.requires(j3)
    return diamond

def diamond_scheduler(watch, duration, msg, scheduler_class=Scheduler):
    """
    create a small diamond scheduler
    total duration = duration
    """
    top = Job(co_print_sleep(watch, duration/4, f"top {msg}"),
              label=f"top {msg}1")
    left = Job(co_print_sleep(watch, duration/2, f"left {msg}"),
               label=f"left {msg}")
    right = Job(co_print_sleep(watch, duration/2, f"right {msg}", ),
                label=f"right {msg}")
    bottom = Job(co_print_sleep(watch, duration / 4, f"bottom {msg}"),
                label=f"bottom {msg}")
    return diamond_from_jobs(watch, top, left, right, bottom)


def diamond_scheduler(watch, duration, msg, scheduler_class=Scheduler):
    """
    create a small diamond scheduler
    total duration = duration
    """
    diamond = scheduler_class(watch=watch)
    top = Job(co_print_sleep(watch, duration/4, f"top {msg}"),
              label=f"top {msg}1",
              scheduler=diamond)
    left = Job(co_print_sleep(watch, duration/2, f"left {msg}"),
               label=f"left {msg}",
               required=top, scheduler=diamond)
    right = Job(co_print_sleep(watch, duration/2, f"right {msg}", ),
                label=f"right {msg}",
                required=top, scheduler=diamond)
    Job(co_print_sleep(watch, duration / 4, f"bottom {msg}"),
        label=f"bottom {msg}",
        required=(left, right), scheduler=diamond)
    return diamond


def pipes(watch, duration, msg, *,
          nb_pipes=2, scheduler_class=Scheduler):
    """
    2 pipes of 2 jobs each
    total duration = duration
    """
    sched = scheduler_class(watch=watch)
    for i in range(1, nb_pipes+1):
        top = Job(co_print_sleep(watch, duration/2, f"top{i} {msg}"),
                  label=f"top{i} {msg}")
        bottom = Job(co_print_sleep(watch, duration/2, f"bot{i} {msg}"),
                     label=f"bot{i} {msg}",
                     required=top)
        sched.update({top, bottom})
    return sched



# pylint: disable=c0111, c0103, c0330, r0201

import unittest

from asynciojobs import Scheduler, Job, Sequence

from asynciojobs import Watch

from .util import co_print_sleep, produce_svg, diamond_scheduler, pipes


class Tests(unittest.TestCase):

    def test_png_easy1(self):
        """
        start with an easy one, a sequence that has a diamond inside
        """
        watch = Watch()
        sched = Scheduler(
            Sequence(
                Job(co_print_sleep(watch, .2, "beg"),
                    label="test_easy"),
                diamond_scheduler(watch, .6, "middle"),
                Job(co_print_sleep(watch, .2, "end")),
            ),
            watch=watch
        )
        produce_svg(sched, "graphic-png-easy1")
        # avoid warning about non-awaited coroutines
        sched.close_idle_jobs()

    def test_png_easy2(self):
        """
        same but with a multi-entry/multi-exit sub-sched
        """
        watch = Watch()
        sched = Scheduler(
            Sequence(
                Job(co_print_sleep(watch, .2, "beg"),
                    label="test_png_easy2"),
                pipes(watch, .6, "middle"),
                Job(co_print_sleep(watch, .2, "end")),
            ),
            watch=watch
        )
        produce_svg(sched, "graphic-png-easy2")
        # avoid warning about non-awaited coroutines
        sched.close_idle_jobs()

    def test_png_simple(self):
        """
        a sequence that has 2 subscheds in a row
        """
        watch = Watch()
        sched = Scheduler(
            Sequence(
                Job(co_print_sleep(watch, .2, "beg"),
                    label="test_png_simple start"),
                diamond_scheduler(watch, .6, "middle-up"),
                pipes(watch, .6, "middle-down"),
                Job(co_print_sleep(watch, .2, "end"),
                    label="test_png_simple end"),
            ),
            watch=watch
        )
        produce_svg(sched, "graphic-png-simple")
        # avoid warning about non-awaited coroutines
        sched.close_idle_jobs()

    def test_png_styles1(self):
        """
        trying the rendering of critical and forever jobs
        """
        watch = Watch()
        sched = Scheduler(
            Sequence(
                Job(co_print_sleep(watch, .1, "regular"),
                    label="regular",
                    critical=False, forever=False),
                Job(co_print_sleep(watch, .1, "critical"),
                    label="critical",
                    critical=True, forever=False),
                Job(co_print_sleep(watch, .1, "forever"),
                    label="forever",
                    critical=False, forever=True),
                Job(co_print_sleep(watch, .1, "both"),
                    label="both",
                    critical=True, forever=True),
            ),
            watch=watch,
        )

        produce_svg(sched, "graphic-png-styles1")
        # avoid warning about non-awaited coroutines
        sched.close_idle_jobs()

    def test_png_styles2(self):
        """
        trying the rendering of critical and forever jobs
        """
        watch = Watch()

        j1 = pipes(watch, .5, "none", nb_pipes=6)
        j1.critical = False
        j1.forever = False
        j1.label = "label-none"

        j2 = diamond_scheduler(watch, .5, "critical")
        j2.critical = True
        j2.forever = False
        j2.label = "label-critical"

        j3 = diamond_scheduler(watch, .5, "forever")
        j3.critical = False
        j3.forever = True
        j3.label = "label-forever"

        j4 = diamond_scheduler(watch, .5, "both")
        j4.critical = True
        j4.forever = True
        j4.label = "label-both"

        sched = Scheduler(
            Sequence(j1, j2, j3, j4),
            watch=watch,
        )

        produce_svg(sched, "graphic-png-styles2")
        # avoid warning about non-awaited coroutines
        sched.close_idle_jobs()


    def test_order1(self):

        async def aprint(x):
            print(x)

        def job(n):
            return Job(aprint(n), label=n)

        sub1, sub2, sub3, sub4 = Scheduler(), Scheduler(), Scheduler(), Scheduler()

        sched = Scheduler(
            Sequence(
                job('top'),
                sub1,
                job('middle'),
                sub2,
                sub3,
                sub4))

        for i in range(3):
            sub1.add(job(i+1))
            sub2.add(job(i+4))
            sub3.add(job(i+7))
            sub4.add(job(i+10))

        sub4.add(job(13))

        produce_svg(sched, "graphic-png-order1")
        # avoid warning about non-awaited coroutines
        sched.close_idle_jobs()



# pylint: disable=c0111, c0103, c0330

import unittest

from asynciojobs import PureScheduler, Job, Sequence

from asynciojobs import Scheduler

from asynciojobs import Watch

from .util import co_print_sleep, produce_svg, diamond_scheduler


class BoomError(Exception):
    pass


async def boom(message):
    raise BoomError(message)


class Tests(unittest.TestCase):

    # xxx probably useless
    def test_nesting1(self):
        """
        one main scheduler in sequence
        one job
        one subscheduler that runs 2 jobs in parallel
        one job
        """

        watch = Watch('test_nesting1')
        # sub-scheduler - total approx 1 s
        sub_sched = PureScheduler(watch=watch)
        Job(co_print_sleep(watch, 0.5, "sub short"), scheduler=sub_sched)
        Job(co_print_sleep(watch, 1, "sub longs"), scheduler=sub_sched)

        # main scheduler - total approx 2 s
        main_sched = PureScheduler(watch=watch)
        Sequence(
            Job(co_print_sleep(watch, 0.5, "main begin")),
            # this is where the subscheduler is merged
            Job(sub_sched.co_run(), label='subscheduler'),
            Job(co_print_sleep(watch, 0.5, "main end")),
            scheduler=main_sched
        )

        print("===== test_nesting1", "LIST with details")
        main_sched.list(details=True)
        ok = main_sched.run()
        self.assertTrue(ok)

        # allow for a small variation around 2s of course
        duration = watch.seconds()
        self.assertAlmostEqual(duration, 2, delta=0.05)

    def test_nesting2(self):
        """
        2 sub schedulers run in parallel while
        the third main one controls them both
        """
        watch = Watch('test_nesting2')
        # sub-scheduler - total approx 0.5 s
        sub2 = diamond_scheduler(watch, 0.5, "SUB2",
                                 scheduler_class=PureScheduler)
        sub2.watch = watch
        # sub-scheduler - total approx 1 s
        sub3 = diamond_scheduler(watch, 1, "SUB3",
                                 scheduler_class=PureScheduler)
        sub3.watch = watch

        # main scheduler - total approx
        # 0.5 + max(0.5, 1) + 0.5 = 2 s
        expected_duration = 2
        main_sched = PureScheduler(watch=watch)
        mainj1 = Job(co_print_sleep(watch, 0.5, "mainj1"), label="mainj1",
                     scheduler=main_sched)
        mainj2 = Job(sub2.co_run(), label="mainj2",
                     required=mainj1,
                     scheduler=main_sched)
        mainj3 = Job(sub3.co_run(), label="mainj3",
                     required=mainj1,
                     scheduler=main_sched)
        Job(co_print_sleep(watch, 0.5, "mainj4"), label="mainj4",
            required=(mainj2, mainj3),
            scheduler=main_sched)

        ok = main_sched.run()
        self.assertTrue(ok)

        # allow for a small variation around 2s of course
        duration = watch.seconds()
        self.assertAlmostEqual(duration, expected_duration, delta=0.05)

    def test_nesting3(self):
        """
        same as test_nesting2
        but using a Scheduler instance
        2 sub schedulers run in parallel while
        the third main one controls them both
        """

        # main scheduler - total approx
        # 0.5 + max(0.5, 1) + 0.5 = 2 s
        expected_duration = 2

        watch = Watch('test_nesting3')
        main_sched = PureScheduler(verbose=True, watch=watch)
        main_sched.label = "main3"
        mainj1 = Job(co_print_sleep(watch, 0.5, "mainj1"), label="mainj1",
                     scheduler=main_sched)

        # sub-scheduler 2 - total approx 0.5 s
        sub_sched2 = diamond_scheduler(watch, 0.5, "SUB2")
        main_sched.add(sub_sched2)
        sub_sched2.requires(mainj1)
        sub_sched2.label = "sub_sched2"
        sub_sched2.verbose = True

        # sub-scheduler 3 - total approx 1 s
        sub_sched3 = diamond_scheduler(watch, 1, "SUB3")
        main_sched.add(sub_sched3)
        sub_sched3.requires(mainj1)
        sub_sched3.label = "sub_sched3"
        sub_sched3.verbose = True

        # last job in main scheduler
        Job(co_print_sleep(watch, 0.5, "mainj4"), label="mainj4",
            required=(sub_sched2, sub_sched3),
            scheduler=main_sched)

        for s in main_sched, sub_sched2, sub_sched3:
            if not s.sanitize():
                print(f"OOPS, had to sanitize sched {s.label}")

        print("===== test_nesting3", "LIST without details")
        main_sched.list(details=False)
        produce_svg(main_sched, "graphic-nesting3")

        watch.reset()
        print("---run")
        ok = main_sched.run()
        if not ok:
            main_sched.debrief()
        self.assertTrue(ok)

        # allow for a small variation around 2s of course
        duration = watch.seconds()
        self.assertAlmostEqual(duration, expected_duration, delta=0.05)

    def test_nesting_sequence(self):

        expected_duration = 1.

        watch = Watch('test_nesting_sequence')

        subjob = Scheduler(
            Sequence(
                Job(co_print_sleep(watch, .2, "one")),
                Job(co_print_sleep(watch, .2, "two")),
                Job(co_print_sleep(watch, .2, "three")),
            ),
            watch=watch,
            label="sub-scheduler\non several lines",
            critical=True, forever=True,
        )

        main = Scheduler(
            Sequence(
                Job(co_print_sleep(watch, .2, "BEGIN"),
                    label="job-label"),
                subjob,
                Job(co_print_sleep(watch, .2, "END")),
            ),
            watch=watch)

        print("===== test_nesting_sequence", "LIST with details")
        main.list(details=True)

        self.assertTrue(main.run())
        self.assertAlmostEqual(watch.seconds(), expected_duration, delta=.05)

        produce_svg(main, "graphic-nesting-sequence")

    def test_critical_exc(self):

        def sched_boom(s_crit, j_crit):
            return Scheduler(
                    Job(boom(str(j_crit)), critical=j_crit),
                    critical=s_crit)

        # regular non-critical schedulers should not raise anything
        # returns True as Job is not critical
        sched = sched_boom(False, False)
        self.assertTrue(sched.run())
        sched.close_idle_jobs()
        # returns False as Job is critical
        sched = sched_boom(False, True)
        self.assertFalse(sched.run())
        sched.close_idle_jobs()

        # it's a different business for critical schedulers
        # Job is not critical, so returns True
        sched = sched_boom(True, False)
        self.assertTrue(sched.run())
        sched.close_idle_jobs()

        # Job is not critical, so raise BoomError
        with self.assertRaises(BoomError):
            sched = sched_boom(True, True)
            sched.run()
        sched.close_all_jobs()

    def test_critical_exc2(self):

        def sched_sched_boom(s1_crit, s2_crit, j_crit):
            return Scheduler(
                    Scheduler(
                        Job(boom("ok"), critical=j_crit,
                            label=f"boom {j_crit}"),
                        critical=s2_crit,
                        label=f"internal {s2_crit}"
                        ),
                    critical=s1_crit,
                    label=f"external {s1_crit}")

        # regular non-critical schedulers should not raise anything
        # returns True as Job is not critical
        s = sched_sched_boom(False, False, False)
        self.assertTrue(s.run())
        s = sched_sched_boom(False, False, True)
        self.assertTrue(s.run())
        s = sched_sched_boom(False, True, False)
        self.assertTrue(s.run())
        s = sched_sched_boom(False, True, True)
        self.assertFalse(s.run())

        # it's a different business for critical schedulers
        # Job is not critical, so returns True
        s = sched_sched_boom(True, False, False)
        self.assertTrue(s.run())
        s = sched_sched_boom(True, False, True)
        self.assertTrue(s.run())
        s = sched_sched_boom(True, True, False)
        self.assertTrue(s.run())
        with self.assertRaises(BoomError):
            s = sched_sched_boom(True, True, True)
            s.run()

    def test_sanitize(self):

        watch = Watch()

        def job(i):
            return Job(co_print_sleep(watch, 0.1, f"job{i}"),
                       label=i)

        def simple():
            all = j1, j2, j3, j4, j5 = [job(i) for i in range(1, 6)]
            sched1 = Scheduler(j1, j2, j3, label='top simple')
            j2.requires(j4)
            j3.requires(j5)
            self.assertEqual(len(j2.required), 1)
            self.assertEqual(len(j3.required), 1)
            sched1.sanitize()
            self.assertEqual(len(j2.required), 0)
            self.assertEqual(len(j3.required), 0)

            for j in all:
                j.close()

        def nested():
            all1 = j11, j12, j13, j14, j15 = [job(i) for i in range(11, 16)]
            sched2 = Scheduler(Sequence(j11, j12, j13), label="nested internal")
            j12.requires(j14)
            j13.requires(j15)

            all = j1, j2, j3, j4, j5 = [job(i) for i in range(1, 6)]
            s1 = Scheduler(Sequence(j1, sched2, j3), label="nested top")
            j1.requires(j4)
            j1.requires(j11)

            sched2.requires(j13)

            # j2 not included in sched, untouched
            j2.requires(j1)

            self.assertEqual(len(j12.required), 2)
            self.assertEqual(len(j13.required), 2)
            self.assertEqual(len(j1.required), 2)
            self.assertEqual(len(sched2.required), 2)
            self.assertEqual(len(j3.required), 1)
            s1.sanitize()
            self.assertEqual(len(j12.required), 1)
            self.assertEqual(len(j13.required), 1)
            self.assertEqual(len(j1.required), 0)
            self.assertEqual(len(sched2.required), 1)
            self.assertEqual(len(j3.required), 1)

            for j in (*all, *all1):
                j.close()

        simple()
        nested()



#!/usr/bin/env python

# pylint: disable=c0111, c0103, r1705

"""
tests for the asynciojobs package
"""

import time
import math
import asyncio
import unittest

from asynciojobs import AbstractJob
from asynciojobs import Job as J
from asynciojobs import Sequence as Seq
from asynciojobs import PureScheduler

from asynciojobs import PrintJob

from asynciojobs import Watch
from .util import co_print_sleep, diamond_from_jobs

# pylint: disable=R0904, R0914, R0903, R1710
# pylint: disable=W0106, W0212, W0201, C0103, C0111, C0321


##############################
def ts():
    """
    a time stamp with millisecond
    """
    # apparently this is not supported by strftime ?!?
    cl = time.time()
    ms = int(1000 * (cl - math.floor(cl)))
    return time.strftime("%M-%S-") + "{:03d}".format(ms)


##############################
async def _sl(n, middle, emergency):
    """
_sl(timeout, middle=False) returns a future that specifies an job like this:
* print incoming `->`
* wait for the time out
* print outgoing `<-`
* return the timeout

_sl(timeout, middle=True) returns a future that specifies an job like this:
* print incoming `->`
* wait for half the time out
* print inside `==` - and optionnally raise an exception there
  if `emergency` is set
* wait for the second half of the time out
* print outgoing `<-`
* return the timeout

"""
    print("{} -> sl({})".format(ts(), n))
    if middle:
        await asyncio.sleep(n / 2)
        print("{} == sl({})".format(ts(), n))
        if emergency:
            raise Exception("emergency exit")
        await asyncio.sleep(n / 2)
    else:
        await asyncio.sleep(n)
    print("{} <- sl({})".format(ts(), n))
    return n


def sl(n): return _sl(n, middle=False, emergency=False)


def slm(n): return _sl(n, middle=True, emergency=False)


##############################
class SleepJob(AbstractJob):

    def __init__(self, timeout, middle=False):
        AbstractJob.__init__(self, forever=False)
        self.timeout = timeout
        self.middle = middle

    def text_label(self):
        return "Sleep for {}s".format(self.timeout)

    async def co_run(self):
        result = await _sl(self.timeout, middle=self.middle, emergency=False)
        return result

    async def co_shutdown(self):
        pass


class TickJob(AbstractJob):

    def __init__(self, cycle):
        AbstractJob.__init__(self, forever=True)
        self.cycle = cycle

    def graph_label(self):
        return "Cyclic tick every {}s".format(self.cycle)

    async def co_run(self):
        counter = 1
        while True:
            print("{} -- Tick {} from {}"
                  .format(ts(), counter, self.text_label()))
            counter += 1
            await asyncio.sleep(self.cycle)

    async def co_shutdown(self):
        pass


async def co_exception(n):
    await asyncio.sleep(n)
    raise ValueError(10**6 * n)

####################
# shortcuts
SLJ = SleepJob
TJ = TickJob

common_sep = 40 * '*' + ' '


def check_required_types(scheduler, message):
    wrong = [j for j in scheduler.jobs
             if not isinstance(j, AbstractJob)
             or not hasattr(j, 'required')]
    if wrong:
        print("in {} : {} {} has {}/{} ill-typed jobs"
              .format(message, type(scheduler).__name__,
                      scheduler, len(wrong), len(scheduler.jobs)))
        return False
    return True


def list_sep(scheduler, sep):
    print(sep)
    scheduler.list()
    print(sep)


class Tests(unittest.TestCase):

    def test_empty(self):                               # pylint: disable=r0201
        sched = PureScheduler()
        sched.list()
        sched.list(details=True)
        self.assertTrue(sched.run())
        # avoid warning about non-awaited coroutines
        sched.close_idle_jobs()


    ####################
    def test_cycle(self):
        """a simple loop with 3 jobs - cannot handle that"""
        a1, a2, a3 = J(sl(1.1)), J(sl(1.2)), J(sl(1.3))
        a1.requires(a2)
        a2.requires(a3)
        a3.requires(a1)

        sched = PureScheduler(a1, a2, a3)

        # these lines seem to trigger a nasty message about a coro not being
        # waited
        self.assertFalse(sched.check_cycles())
        # avoid warning about non-awaited coroutines
        sched.close_idle_jobs()

    ####################
    # Job(asyncio.sleep(0.4))
    # or
    # SleepJob(0.4)
    # are almost equivalent forms to do the same thing
    def test_simple(self):
        """a simple topology, that should work"""
        jobs = SLJ(0.1), SLJ(0.2), SLJ(0.3), SLJ(
            0.4), SLJ(0.5), J(sl(0.6)), J(sl(0.7))
        a1, a2, a3, a4, a5, a6, a7 = jobs
        a4.requires(a1, a2, a3)
        a5.requires(a4)
        a6.requires(a4)
        a7.requires(a5)
        a7.requires(a6)

        sched = PureScheduler(*jobs)
        list_sep(sched, common_sep + "LIST BEFORE")
        self.assertTrue(sched.check_cycles())
        self.assertTrue(sched.run())
        for j in jobs:
            self.assertFalse(j.raised_exception())
        list_sep(sched, common_sep + "LIST AFTER")
        print(common_sep + "DEBRIEF")
        sched.debrief()
        print(common_sep + "DEBRIEF and silence jobs that are done")
        sched.debrief(silence_done_jobs=True)

    ####################
    def test_forever1(self):
        a1, a2, t1 = SLJ(1), SLJ(1.5), TJ(.6)
        a2.requires(a1)
        sched = PureScheduler(a1, a2, t1)
        sched.list()
        self.assertTrue(sched.run())
        sched.list()

    ####################
    def test_timeout1(self):
        a1, a2, a3 = [SLJ(x) for x in (0.5, 0.6, 0.7)]
        a2.requires(a1)
        a3.requires(a2)
        sched = PureScheduler(a1, a2, a3, timeout=1)
        # should timeout in the middle of stage 2
        self.assertFalse(sched.run())
        sched.list()

    ####################
    def _test_exc_non_critical(self, verbose):

        print("verbose = {}".format(verbose))
        a1, a2 = SLJ(1), J(co_exception(0.5), label='non critical boom',
                           critical=False)
        sched = PureScheduler(a1, a2, verbose=verbose)
        self.assertTrue(sched.run())
        print(common_sep + 'debrief()')
        sched.debrief()

    def test_exc_non_critical_f(
        self): return self._test_exc_non_critical(False)

    def test_exc_non_critical_t(self): return self._test_exc_non_critical(True)

    ####################
    def _test_exc_critical(self, verbose):

        print("verbose = {}".format(verbose))
        a1, a2 = SLJ(1), J(co_exception(0.5),
                           label='critical boom', critical=True)
        sched = PureScheduler(a1, a2, verbose=verbose)
        self.assertFalse(sched.run())
        print(common_sep + 'debrief()')
        sched.debrief()

    def test_exc_critical_f(self): return self._test_exc_critical(False)

    def test_exc_critical_t(self): return self._test_exc_critical(True)

    ####################
    def test_sequence1(self):
        "a simple sequence"
        a1 = J(sl(0.1), label=1)
        a2 = J(sl(0.1), label=2)
        a3 = J(sl(0.1), label=3)
        s = Seq(a1, a2, a3)
        sched = PureScheduler(s)
        list_sep(sched, common_sep + "sequence1")
        self.assertEqual(len(a1.required), 0)
        self.assertEqual(len(a2.required), 1)
        self.assertEqual(len(a3.required), 1)
        self.assertTrue(check_required_types(sched, "test_sequence1"))
        self.assertTrue(sched.run())

    ####################
    def test_sequence2(self):
        "a job and a sequence"
        a1 = J(sl(0.1), label=1)
        a2 = J(sl(0.1), label=2)
        a3 = J(sl(0.1), label=3)
        s = Seq(a2, a3, required=a1)
        sched = PureScheduler(a1, s)
        list_sep(sched, common_sep + "sequence2")
        self.assertEqual(len(a1.required), 0)
        self.assertEqual(len(a2.required), 1)
        self.assertEqual(len(a3.required), 1)
        self.assertTrue(check_required_types(sched, "test_sequence2"))
        self.assertTrue(sched.run())

    ####################
    def test_sequence3(self):
        "a sequence and a job"
        a1 = J(sl(0.1), label=1)
        a2 = J(sl(0.1), label=2)
        s = Seq(a1, a2)
        a3 = J(sl(0.1), label=3, required=s)
        sched = PureScheduler()
        sched.update((s, a3))
        list_sep(sched, common_sep + "sequence3")
        self.assertEqual(len(a1.required), 0)
        self.assertEqual(len(a2.required), 1)
        self.assertEqual(len(a3.required), 1)
        self.assertTrue(check_required_types(sched, "test_sequence3"))
        self.assertTrue(sched.run())

    ####################
    def test_sequence4(self):
        "a sequence of 2 sequences"
        a1 = J(sl(0.1), label=1)
        a2 = J(sl(0.1), label=2)
        a3 = J(sl(0.1), label=3)
        a4 = J(sl(0.1), label=4)
        s1 = Seq(a1, a2)
        s2 = Seq(a3, a4)
        sched = PureScheduler(Seq(s1, s2))
        list_sep(sched, common_sep + "sequence4")
        self.assertEqual(len(a1.required), 0)
        self.assertEqual(len(a2.required), 1)
        self.assertEqual(len(a3.required), 1)
        self.assertEqual(len(a4.required), 1)
        self.assertTrue(check_required_types(sched, "test_sequence4"))
        self.assertTrue(sched.run())

    ####################
    def test_sequence5(self):
        "sequences with required"
        a1 = J(sl(0.1), label=1)
        a2 = J(sl(0.1), label=2)
        a3 = J(sl(0.1), label=3)
        a4 = J(sl(0.1), label=4)
        a5 = J(sl(0.1), label=5)
        a6 = J(sl(0.1), label=6)
        s1 = Seq(a1, a2)
        s2 = Seq(a3, a4, required=s1)
        s3 = Seq(a5, a6, required=s2)
        sched = PureScheduler(s1, s2, s3)
        list_sep(sched, common_sep + "sequence5")
        self.assertEqual(len(a1.required), 0)
        self.assertEqual(len(a2.required), 1)
        self.assertEqual(len(a3.required), 1)
        self.assertEqual(len(a4.required), 1)
        self.assertEqual(len(a5.required), 1)
        self.assertEqual(len(a6.required), 1)
        self.assertTrue(check_required_types(sched, "test_sequence5"))
        self.assertTrue(sched.run())

    ##########
    def test_sequence6(self):
        "adding a sequence"
        sched = PureScheduler()
        a1 = J(sl(0.1), label=1)
        a2 = J(sl(0.1), label=2)
        a3 = J(sl(0.1), label=3)
        sched.add(Seq(a1, a2, a3))
        self.assertTrue(sched.run())

    ##########
    def test_requires_job(self):
        # the calls to close() are here to avoid
        # the warning about non-awaited coroutines

        a1 = J(sl(0.1), label="a1")
        a2 = J(sl(0.1), label="a2")
        a3 = J(sl(0.1), label="a3")
        a4 = J(sl(0.1), label="a4")
        a5 = J(sl(0.1), label="a5")

        # several forms to create
        b = J(sl(0.2), required=None)
        self.assertEqual(len(b.required), 0)
        b.close()
        b = J(sl(0.2), required=(None,))
        self.assertEqual(len(b.required), 0)
        b.close()
        b = J(sl(0.2), required=[None])
        self.assertEqual(len(b.required), 0)
        b.close()
        b = J(sl(0.2), required=a1)
        self.assertEqual(len(b.required), 1)
        b.close()
        b = J(sl(0.2), required=(a1,))
        self.assertEqual(len(b.required), 1)
        b.close()
        b = J(sl(0.2), required=[a1])
        self.assertEqual(len(b.required), 1)
        b.close()
        b = J(sl(0.2), label='BROKEN', required=(a1, a2))
        self.assertEqual(len(b.required), 2)
        b.close()
        b = J(sl(0.2), required=[a1, a2])
        self.assertEqual(len(b.required), 2)
        b.close()
        b = J(sl(0.2), required=[a1, (a2,), set([a3, a4]), [[[[[[a5]]]]]]])
        self.assertEqual(len(b.required), 5)
        b.close()

        for a in a1, a2, a3, a4, a5:
            a.close()

    ##########
    def test_requires_sequence(self):

        # we leave these untouched (no req. added)
        a1 = J(sl(0.1), label="a1")
        a2 = J(sl(0.1), label="a2")
#        a3 = J(sl(0.1), label="a3")
#        a4 = J(sl(0.1), label="a4")
#        a5 = J(sl(0.1), label="a5")

        # re-create these each time to have fresh data
        def bs():
            b1 = J(sl(0.1), label="b1")
            b2 = J(sl(0.1), label="b2")
            b3 = J(sl(0.1), label="b3")
            return b1, b2, b3

        b1, b2, b3, *_ = bs()
        s1 = Seq(b1, b2, b3, required=[a1, a2])
        self.assertEqual(len(b1.required), 2)
        self.assertEqual(len(b2.required), 1)
        s1.close_idle_jobs()

        b1, b2, b3, *_ = bs()
        s1 = Seq(b1, b2, b3)
        s1.requires([a1, a2])
        self.assertEqual(len(b1.required), 2)
        self.assertEqual(len(b2.required), 1)
        s1.close_idle_jobs()
        
        for a in a1, a2:
            a.close()


    ##########
    def test_timeout2(self):
        a1 = J(sl(1), label="a1")
        a2 = J(sl(2), label="a2")
        a3 = J(sl(10), label="a3")
        result = PureScheduler(a1, a2, a3, timeout=3).run()
        self.assertEqual(result, False)
        self.assertEqual(a1.is_done(), True)
        self.assertEqual(a1.result(), 1)
        self.assertEqual(a2.is_done(), True)
        self.assertEqual(a2.result(), 2)
        self.assertEqual(a3.is_done(), False)

    ##########
    def test_forever2(self):
        async def tick(n):
            while True:
                print('tick {}'.format(n))
                await asyncio.sleep(n)

        a1 = J(sl(0.5), label="finite")
        a2 = J(tick(0.1), forever=True, label="forever")
        sched = PureScheduler(a1, a2)
        result = sched.run()
        self.assertEqual(result, True)
        self.assertEqual(a1.is_done(), True)
        self.assertEqual(a2.is_done(), False)

    ##########
    def test_creation_scheduler(self):
        sched = PureScheduler()
        s = Seq(J(sl(1)), J(sl(2)), scheduler=sched)
        J(sl(3), required=s, scheduler=sched)
        # make sure that jobs appended in the sequence
        # even later on are also added to the scheduler
        s.append(J(sl(.5)))
        self.assertEqual(len(sched.jobs), 4)
        self.assertTrue(sched.check_cycles())
        self.assertTrue(sched.run())

    def test_seq(self):
        s = PureScheduler()
        Seq(J(sl(.1)), J(sl(.2)),
            scheduler=s)
        self.assertTrue(s.run())

    # if window is defined, total should be a multiple of window
    def _test_window(self, total, window):
        atom = .1
        tolerance = 8  # more or less % in terms of overall time
        s = PureScheduler(jobs_window=window)
        for i in range(1, total + 1):
            s.add(PrintJob("{}-th {}s job".
                           format(i, atom),
                           sleep=atom))
        beg = time.time()
        ok = s.run()
        ok or s.debrief(details=True)
        end = time.time()
        duration = end - beg

        # estimate global time
        # unwindowed: overall duration is atom
        # otherwise a multiple of it (assuming total = k*window)
        expected = atom if not window else (total / window) * atom
        print('overall expected {} - measured {}'
              .format(expected, duration))

        distortion = duration / expected
        time_ok = 1 - tolerance / 100 <= distortion <= 1 + tolerance / 100
        if not time_ok:
            print("_test_window - window = {} :"
                  "wrong execution time {} - not within {}% of {}"
                  .format(window, end - beg, tolerance, expected))

        self.assertTrue(time_ok)
        self.assertTrue(ok)

    def test_window(self):
        self._test_window(total=15, window=3)

    def test_no_window(self):
        self._test_window(total=15, window=None)

    ##########
    def test_display(self):

        class FakeTask:

            def __init__(self):
                self._result = 0
                self._exception = None

        def annotate_job_with_fake_task(job, state, boom):
            task = FakeTask()
            if state == "done":
                task._state = asyncio.futures._FINISHED
                job._task = task
                job._running = True
            elif state == "running":
                task._state = "NONE"
                job._task = task
                job._running = True
            elif state == "scheduled":
                task._state = "NONE"
                job._task = task
                job._running = False
            else:
                pass

            # here we assume that a job that has raised an exception is
            # necessarily done
            if boom:
                if state in ("idle", "scheduled", "running"):
                    print("incompatible combination boom x idle - ignored")
                    return
                else:
                    job._task._exception = True
            return job

        class AJ(AbstractJob):
            pass

        sched = PureScheduler()
        previous = None
        for state in "idle", "scheduled", "running", "done":
            for boom in True, False:
                for critical in True, False:
                    for forever in True, False:
                        j = AJ(critical=critical,
                               forever=forever,
                               label="forever={} crit.={} status={} boom={}"
                               .format(forever, critical, state, boom),
                               required=previous)
                        if annotate_job_with_fake_task(j, state, boom):
                            sched.add(j)
                            previous = j
        sched.list()


    def test_iterate(self):
        watch = Watch()
        sched = diamond_from_jobs(
            watch,
            J(co_print_sleep(watch, .1, "1")),
            diamond_from_jobs(
                watch,
                J(co_print_sleep(watch, .1, "2.1")),
                J(co_print_sleep(watch, .1, "2.2")),
                J(co_print_sleep(watch, .1, "2.3")),
                J(co_print_sleep(watch, .1, "2.4")),
                ),
            diamond_from_jobs(
                watch,
                J(co_print_sleep(watch, .1, "3.1")),
                J(co_print_sleep(watch, .1, "3.2")),
                J(co_print_sleep(watch, .1, "3.3")),
                J(co_print_sleep(watch, .1, "3.4")),
                ),
            J(co_print_sleep(watch, .1, "4")))

        c1 = sum(1 for job in sched.iterate_jobs(scan_schedulers=False))
        c2 = sum(1 for job in sched.iterate_jobs(scan_schedulers=True))

        # without schedulers we expect 10 jobs
        self.assertEqual(c1, 10)
        # with schedulers we expect 13 jobs since top scheduler
        # gets taken into account too
        self.assertEqual(c2, 13)
        # avoid warning about non-awaited coroutines
        sched.close_idle_jobs()


if __name__ == '__main__':
    import sys
    import asynciojobs
    if '-v' in sys.argv:
        # set module flag
        asynciojobs.purescheduler.debug = True
        sys.argv.remove('-v')
    unittest.main()
